from typing import Sequence


def join_with_commas_and_and(items: Sequence[str]) -> str:
    if len(items) == 0:
        return ''
    elif len(items) == 1:
        return items[0]
    elif len(items) == 2:
        # Use only and, not commas
        return f'{items[0]} and {items[1]}'

    # Length 3+, handled the same, commas for all and also and on the last join
    out_str = ', '.join(items[:-1])
    out_str += f', and {items[-1]}'
    return out_str



from typing import Sequence, Any
from pyexlatex.presentation.beamer.frame.frame import Frame
from pyexlatex.models.format.adjustbox import AdjustBox
from pyexlatex.models.sizes.textwidth import TextWidth
from pyexlatex.models.sizes.textheight import TextHeight
from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models.format.fills import VFill
from pyexlatex.models.format.centering import Center


class FullWidthFrame(Frame):
    """
    Resizes passed latex object to take up entire frame.
    """

    def __init__(self, content: Sequence[Sequence[Any]], **kwargs):
        content = adjust_to_full_size(content)
        super().__init__(content, **kwargs)


class GraphicFrame(FullWidthFrame):
    """
    Resizes passed graphic to take up entire frame. Can pass a file path or a latex object.
    """

    def __init__(self, content: Any, **kwargs):
        if isinstance(content, str):
            content = Graphic(content)
        super().__init__(content, **kwargs)


class MultiGraphicFrame(Frame):
    """
    Resizes each graphic to full width and puts vertical space in between graphics.
    Can pass a file path or a latex object.
    """

    def __init__(self, content: Sequence[Sequence[Any]], **kwargs):
        all_content = []
        for cont in content:
            if isinstance(content, str):
                cont = Graphic(cont)
            cont = adjust_to_full_size(cont)
            all_content.extend([cont, VFill()])
        all_content = all_content[:-1]  # strip final VFill
        super().__init__(all_content, **kwargs)


def adjust_to_full_size(content: Any) -> Center:
    """
    Takes content and adjusts it to 90% text width, 80% text height, but keeping the aspect ratio.
    """

    content = AdjustBox(
        content,
        adjust_options=[
            f'width=0.9{TextWidth()}',
            f'height=0.8{TextHeight()}',
            'keepaspectratio'

        ]
    )
    content = Center(content)
    return content



from pyexlatex.models.item import MultiOptionSimpleItem

class MultiColumn(MultiOptionSimpleItem):
    name = 'multicolumn'

    def __init__(self, contents: str, span: int, align: str='c'):
        self.span = span
        super().__init__(self.name, span, align, contents)

    def __len__(self):
        return self.span



from pyexlatex.table.models.panels import Panel
from pyexlatex.table.models.data.table import DataTable
from pyexlatex.table.models.labels.table import LabelCollection, LabelTable
from pyexlatex.table.models.labels.label import Label
from pyexlatex.table.models.table.table import Table


from typing import Optional
from copy import deepcopy
from pyexlatex.models import Item
from pyexlatex.models.section.base import TextAreaBase
from pyexlatex.presentation.beamer.frame.title import FrameTitle


class Frame(TextAreaBase):
    """
    Basic unit for creating a slide in a presentation.
    """
    name = 'frame'

    def __init__(self, content, title: Optional[str] = None, label: Optional[str] = None, **kwargs):
        self.title = title
        if isinstance(content, (Item, str)):
            content = [content]
        else:
            content = deepcopy(content)  # don't modify content inplace

        if self.title is not None:
            content.insert(0, FrameTitle(title))

        self.content = content
        # TODO: evaluate whether fragile can be passed for all frames and what the performance hit is. If cannot, then
        # TODO: must determine from contents whether fragile should be passed. Will require adding boolean data types
        # TODO: to item data
        super().__init__(self.name, self.content, label=label, env_modifiers='[fragile]', **kwargs)


from pyexlatex.models.item import SimpleItem


class VSpace(SimpleItem):
    """
    Manually insert vertical spacing, of a given height.
    """
    name = 'vspace'

    def __init__(self, height: float = 0.5):
        super().__init__(self.name, contents=f'{height}cm')



from typing import List, Optional, Sequence, Union
from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.models.package import Package
from mixins.repr import ReprMixin
from mixins.attrequals import EqOnAttrsMixin, EqHashMixin
from pyexlatex.models.references.bibtex.base import BibTexEntryBase


class UniqueDataList(list):
    """
    A list where the items are always unique.
    """

    def __init__(self, iterable):
        if iterable is None:
            iterable = []
        super().__init__(iterable)

    def append(self, value):
        if value is None:
            return
        if value in self:
            return
        return super().append(value)

    def extend(self, iterable):
        if iterable is None:
            return
        new_items = [item for item in iterable if item not in self]
        if not new_items:
            return
        return super().extend(new_items)


class UniquePackagesList(UniqueDataList):
    """
    A unique data list that allows passing of just a string to represent a package
    """

    def __init__(self, packages: Optional[List[Union[Package, str]]] = None):
        if packages is None:
            packages = []

        packages = self._as_packages(packages)

        super().__init__(packages)

    def _as_package(self, package: Union[Package, str]):
        if isinstance(package, str):
            return Package(package)

        return package

    def _as_packages(self, packages: Optional[List[Union[Package, str]]] = None):
        return [self._as_package(pack) for pack in packages]

    def append(self, package: Union[Package, str]):
        if package is None:
            return
        package = self._as_package(package)
        super().append(package)

    def extend(self, packages: Optional[List[Union[Package, str]]] = None):
        if packages is None:
            return
        packages = self._as_packages(packages)
        super().extend(packages)


class DocumentSetupData(ReprMixin, EqOnAttrsMixin, EqHashMixin):
    repr_cols = [
        'filepaths',
        'begin_document_items',
        'end_document_items',
        'packages',
        'source_paths',
        'references'
    ]
    ignore_attrs = [
        'ignore_attrs',
        'repr_cols',
        'attrs',
        'readable_repr',
        'equal_attrs'
    ]
    equal_attrs = [
        'filepaths',
        'binaries',
        'begin_document_items',
        'end_document_items',
        'packages',
        'source_paths',
        'references'
    ]

    def __init__(self, filepaths: Optional[List[str]] = None, binaries: Optional[List[bytes]] = None, 
                 begin_document_items: Optional[List[DocumentItem]] = None, 
                 end_document_items: Optional[List[DocumentItem]] = None, packages: Optional[List[Package]] = None,
                 source_paths: Optional[List[str]] = None, references: Optional[Sequence[BibTexEntryBase]] = None):
        self.filepaths = UniqueDataList(filepaths)
        self.binaries = UniqueDataList(binaries)
        self.begin_document_items = UniqueDataList(begin_document_items)
        self.end_document_items = UniqueDataList(end_document_items)
        self.packages = UniquePackagesList(packages)
        self.source_paths = UniqueDataList(source_paths)
        self.references = UniqueDataList(references)

    @property
    def attrs(self) -> List[str]:
        return [attr for attr in dir(self) if (not attr.startswith('_') and not attr in self.ignore_attrs)]


from typing import Optional, Sequence, Union
from pyexlatex.models.item import Item
from pyexlatex.models.containeritem import ContainerItem
from mixins.repr import ReprMixin
from pyexlatex.table.models.panels.collection import PanelCollection
from pyexlatex.table.models.texgen.alignment import ColumnsAlignment
from pyexlatex.table.logic.table.build import build_tabular_content_from_panel_collection
from pyexlatex.models.caption import Caption
from pyexlatex.models.format.breaks import LineBreak
from pyexlatex.texgen import _centering_str
from pyexlatex.models.document import Document
from pyexlatex.models.package import Package
from pyexlatex.table.models.texgen.packages import default_packages
from pyexlatex.models.landscape import Landscape
from pyexlatex.models.label import Label
from pyexlatex.models.section.base import TextAreaBase


class TableNotes(TextAreaBase, ReprMixin):
    name = 'tablenotes'

    def __init__(self, contents: Union[str, Sequence[str]]):
        super().__init__(self.name, contents, env_modifiers=f'[para, flushleft]')

class Tabular(Item, ReprMixin):
    name = 'tabular'
    repr_cols = ['align']

    def __init__(self, panel_collection: PanelCollection, align: ColumnsAlignment=None,
                 mid_rules=True):
        self.align = align if align is not None else ColumnsAlignment(num_columns=panel_collection.num_columns)
        self.panel_collection = panel_collection

        content = build_tabular_content_from_panel_collection(panel_collection, mid_rule=mid_rules)

        super().__init__(self.name, content, env_modifiers=f'{{{self.align}}}')

class ThreePartTable(Item, ReprMixin):
    name = 'threeparttable'
    repr_cols = ['caption']

    def __init__(self, table_content: Tabular, caption: Caption=None, below_text: TableNotes=None,
                 label: Optional[Label] = None):
        self.caption = caption
        items = [
            caption,
            table_content,
            below_text,
            label
        ]
        valid_items = [item for item in items if item is not None]

        content = LineBreak().join(valid_items)
        super().__init__(self.name, content)

    @classmethod
    def from_panel_collection(cls, panel_collection: PanelCollection, *args, tabular_kwargs={}, **kwargs):
        tabular = Tabular(panel_collection, **tabular_kwargs)

        if panel_collection.name is not None:
            caption = Caption(panel_collection.name)
        else:
            caption = None

        return cls(tabular, caption=caption, *args, **kwargs)

class Table(ContainerItem, Item, ReprMixin):
    name = 'table'
    repr_cols = ['caption']

    def __init__(self, three_part_table: ThreePartTable, centering=True, landscape=False):
        self.caption = three_part_table.caption
        self.landscape = landscape

        items = [
            _centering_str() if centering else None,
            three_part_table
        ]

        valid_items = [item for item in items if item is not None]

        content = LineBreak().join(valid_items)

        Item.__init__(self, self.name, content)

    def __str__(self):
        content_with_env = super().__str__()
        if self.landscape:
            content_with_env = Landscape().wrap(str(content_with_env))
        return content_with_env

    @classmethod
    def from_panel_collection(cls, panel_collection: PanelCollection, *args, tabular_kwargs={},
                              three_part_table_kwargs={}, **kwargs):
        three_part_table = ThreePartTable.from_panel_collection(
            panel_collection,
            tabular_kwargs=tabular_kwargs,
            **three_part_table_kwargs
        )
        return cls(three_part_table, *args, **kwargs)

    @classmethod
    def from_table_model(cls, table, *args, **kwargs):
        from pyexlatex.table.models.table.table import Table as TableModel
        table: TableModel
        tabular = Tabular(
            table.panels,
            align=table.align,
            mid_rules=table.mid_rules
        )

        three_part_table = ThreePartTable(
            tabular,
            caption=table.caption,
            below_text=table.below_text,
            label=Label(table.label) if table.label else None
        )
        obj = cls(three_part_table, *args, landscape=table.landscape, **kwargs)
        obj.add_data_from_content(table)
        return obj



class LTable(Table):
    name = 'ltable'


class TableDocument(Document):

    def __init__(self, content: Table, packages: [Package]=None, landscape: bool=False):
        if packages is None:
            packages = []

        packages += default_packages

        super().__init__(content, packages, landscape=landscape)

    @classmethod
    def from_panel_collection(cls, panel_collection: PanelCollection, *args, tabular_kwargs={},
                              three_part_table_kwargs={}, table_kwargs={}, **kwargs):
        table = Table.from_panel_collection(
            panel_collection,
            tabular_kwargs=tabular_kwargs,
            three_part_table_kwargs=three_part_table_kwargs,
            **table_kwargs
        )
        return cls(table, *args, **kwargs)

    @classmethod
    def from_table_model(cls, table, *args, **kwargs):
        from pyexlatex.table.models.table.table import Table as TableModel
        table: TableModel
        tex_table = Table.from_table_model(table, *args, **kwargs)
        obj = cls(tex_table, *args, **kwargs)
        obj.add_data_from_content(table)
        return obj



from typing import List
from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.models.title.title import Title
from pyexlatex.models.credits.author import Author
from pyexlatex.models.date import Date
from pyexlatex.models.section.abstract import Abstract
from pyexlatex.texgen import _maketitle_str
from pyexlatex.logic.format.and_join import join_with_commas_and_and

class TitlePage(DocumentItem):

    def __init__(self, title: str = None, authors: List[str] = None, date: str = None, abstract: str = None):
        from pyexlatex.logic.builder import _build

        author = join_with_commas_and_and(authors)

        contents = [
            Title(title) if title is not None else None,
            Author(author, short_author=None) if author is not None else None,
            Date(date) if date is not None else Date(),
            _maketitle_str() if title is not None else None,
            Abstract(abstract) if abstract is not None else None
        ]

        self.contents = [content for content in contents if content is not None]

        self._output = _build(self.contents)

    def __str__(self):
        return self._output


from pyexlatex.models.item import SimpleItem


class Ref(SimpleItem):
    """
    Pass a label to create an in text reference to that label.
    """
    name = 'ref'

    def __init__(self, contents):
        super().__init__(self.name, contents)


from typing import Optional, Sequence, Union
import pyexlatex as pl
from pyexlatex.logic.format.and_join import join_with_commas_and_and
from pyexlatex.models.format.hangindent import HangIndent


class Publication(pl.Template):

    def __init__(self, title: str, co_authors: Optional[Sequence[str]] = None, journal_info: Optional[str] = None,
                 href: Optional[str] = None, extra_info: Optional[str] = None):
        self.title = title
        self.co_authors = co_authors
        self.journal_info = journal_info
        self.href = href
        self.extra_info = extra_info

        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        possible_items = [
            HangIndent(1),
            self._title_str,
            pl.Italics(self.journal_info) if self.journal_info is not None else None,
            self._with_coauthors,
            self.extra_info
        ]
        items = [item for item in possible_items if item is not None]
        return items

    @property
    def _title_obj(self) -> Union[pl.Bold, pl.Hyperlink]:
        if self.href is None:
            title = pl.Bold(self.title)
        else:
            # Handle hyperlink
            styled_title = pl.TextColor(self.title, color=pl.RGB(50, 82, 209, color_name='darkblue'))
            styled_title = pl.Underline(styled_title)
            title = pl.Hyperlink(self.href, styled_title)
            self.add_data_from_content(title)
        return title

    @property
    def _title_str(self) -> str:
        title_str = str(self._title_obj)
        if any([item is not None for item in [self.co_authors, self.journal_info, self.extra_info]]):
            title_str += ','
        return title_str


    @property
    def _with_coauthors(self) -> Optional[str]:
        if self.co_authors is None:
            return None

        return f' with {join_with_commas_and_and(self.co_authors)}'
